/*
	utility program for the ProWesS loader

	COPY : Copy one or more files, or even entire directory trees,
	the file versions and update and backup dates are retained when
	copying.

	The parameters are all the files and or directories which have to be
	copied, possibly some flags, and as last parameter the destination
	directory (filenames are always retained).

	If an error occurs while copying, then the error message is returned.
	Files which are only partly copied will then be deleted.

	possible flags :
	-tree : copy the entire subdirectory and create the subdirectories
		if they did not exist
	-move : delete all the files after copying
	-backup : only copy files if the distination file (if exists) is older than
		the source
	-progres, -progress : display progres report
	-new : only copy new files, do not overwrite existing files

	written by Joachim Van der Auwera for PROGS
	(c) 1996 by PROGS, Professional & Graphical Software
	started 17 April 1996
	23.07.1996 sometimes ERR_MCHK is done in IOFileLoad but not in IOGetBytes
			   so use slower method if fast one fails
	02.09.1996 -new option added
	04.03.1997 -new should now function properly
*/

#include "str.h"
#include "mem.h"
#include "dir.h"
#include "io.h"
#include "job.h"
#include "ProWesS.h"

#define catch(x) do { if (err=(x)) return err; } while (0)

#define MAX_MESSAGE 50

typedef struct {
	char **params;					/* parameters */
	char *dest; 					/* destination directory */
	char tree;						/* copy entire directory trees */
	char move;						/* move files */
	char backup;					/* backup files */
	char new;						/* only new files */
	char progres;					/* display progres of copying */
	Size bufflen;					/* length of buffer for copying */
	char *buffer;					/* base of buffer or NULL */
	Job subjob; 					/* subjob for progres report window */
	char display[MAX_MESSAGE];		/* progres report message */
	PWObject message;				/* object with message */
	char lastcopied[IO_MAXFULLNAME];/* last copied file */
	} Global;

Error copy(Global *g, char *filename);
Error progres(Global *g);

Error init(char *param)
{
	Error err;
	char **params;
	int count=-1;					/* no of files to copy */
	Global g;

	MEMClear(sizeof(Global),&g);
	g.display[0]=' ';

	/* parse the command line parameters */
	catch( STRBreak(param,&g.params) );
	for (params=g.params ; *params ; params++)
	{
		if (STRSameCI("-path",*params))
		{
			if (!params[1]) return ERR_IPAR;
			*params++=NULL;
			*params  =NULL;
		}
		else
		if (STRSameCI("-tree",*params))
		{
			g.tree=TRUE;
			*params=NULL;
		}
		else
		if (STRSameCI("-move",*params))
		{
			g.move=TRUE;
			*params=NULL;
		}
		else
		if (STRSameCI("-backup",*params))
		{
			g.backup=TRUE;
			*params=NULL;
		}
		else
		if (STRSameCI("-progres",*params) || STRSameCI("-progress",*params))
		{
			g.progres=TRUE;
			*params=NULL;
		}
		else
		if (STRSameCI("-new",*params))
		{
			g.new=TRUE;
			*params=NULL;
		}
		else
		{
			count++;
			g.dest=*params;
		}
	}

	/* need something to copy */
	if (count<1) return ERR_IPAR;
	if (g.new) g.backup=FALSE;

	/* display progress report window */
	if (g.progres)
	{
		g.progres=FALSE;
		JOBCall(progres,"copy",64,DEFAULT_STACK,&g.subjob,1,&g);
	}

	/* params should indicate first file/directory to copy */
	params=g.params;
	while (!*params) params++;

	for ( ; count>0 ; params++, count--)
	{
		while (!*params) params++;
		err=copy(&g,*params);
		if (err)
		{
			/* delete destination */
			IODelete(g.lastcopied);
			return err;
		}
	}
	return ERR_OK;
}

Error filecopy(Global *g, Channel in, FileInfo *info)
{
	Channel out;
	Error err;
	char name[IO_MAXNAME];
	Size done, todo;

	/* open the output file */
	if (g->backup)
	{
		err=IOOpenPath(info->name,OPEN_OLD,g->dest,&out);
		if (!err)
		{
			Time old;
			IOUpdateTimeGet(out,&old);
			if (info->UpdateTime<=old)
			{
				IOClose(out);
				IOClose(in);
				return err;
			}
			IOPositionSet(out,0);
			IOTruncate(out);
		}
		else
		{
			err=IOOpenPath(info->name,OPEN_OVER,g->dest,&out);
			if (err) { IOClose(in); return err; }
		}
	}
	else
	{
		err=IOOpenPath(info->name,g->new?OPEN_NEW:OPEN_OVER,g->dest,&out);
		if (err) { IOClose(in); return (err==ERR_FEX && g->new)?ERR_OK:err; }
	}

	if (g->progres) JOBEventSend(g->subjob,EVENT_LEVEL1);

	/* fill in destination filename for deletion on error */
	err=IOName(out,name,g->lastcopied);
	if (err) { IOClose(out); IOClose(in); return err; }
	STRAppend(name,g->lastcopied);

	/* allocate the file buffer */
	if (info->length>g->bufflen)
	{
		Size size=(info->length+0xfff)&0xfffff000;

		MEMRelease(g->buffer);
		err=MEMAllocate(size,&g->buffer);
		while (err)
		{
			size=size/2;
			err=MEMAllocate(size,&g->buffer);
		}
		g->bufflen=size;
	}

	/* read and write all the data */
	if (info->length<=g->bufflen)
	{
		err=IOFileLoad(in,info->length,g->buffer);
		if (err)
		{
			/* if medium check, try again the slow way */
			/* this might be more accurate (proven in some tests) */
			if (err==ERR_MCHK) goto slow;
			IOClose(out); IOClose(in); return err;
		}
		err=IOFileSave(out,info->length,g->buffer);
		if (err) { IOClose(out); IOClose(in); return err; }
	}
	else
	{
slow:	for (todo=info->length ; todo>0 ; todo-=done)
		{
			err=IOGetBytes(in,min(todo,g->bufflen),g->buffer,&done);
			if (err) { IOClose(out); IOClose(in); return err; }
			err=IOPutBytes(out,done,g->buffer,NULL);
			if (err) { IOClose(out); IOClose(in); return err; }
		}
	}

	/* fix the header */
	IOFileInfoSet(out,info);

	/* copying this file was succesful, so should not be deleted */
	g->lastcopied[0]='\0';

	if (g->move)
	{
		char dir[IO_MAXFULLNAME];

		IOName(in,name,dir);
		IOClose(in);
		STRAppend(name,dir);
		IODelete(dir);
	}

	/* done */
	IOClose(out);
	IOClose(in);
	return err;
}

Error treecopy(Global *g, char *dirname, FileInfo *info)
{
	Error err;
	Channel in;
	DirIterator it;
	char dir[IO_MAXDIRECTORY];
	char name[IO_MAXFULLNAME];

	/* in case of tree, build the directory subtree */
	if (g->tree)
	{
		catch( DIRSubOpen(dirname,TRUE,&it) );
		while (!DIRSubNext(it,info))
		{
			STRCopy(g->dest,name);
			if (name[strlen(name)-1]!='_') STRAppendSafe("_",name,IO_MAXFULLNAME);
			STRAppendSafe(info->name,name,IO_MAXFULLNAME);
			err=IOMakeDirectory(name);
			/* no problem if directory already existed */
			if (err && err!=ERR_FEX) return err;
		}
		DIRSubClose(it);
	}

	catch( DIROpen(dirname,g->tree,&it,NULL) );
	catch( DIRDirectory(it,dir) );
	while (!DIRNext(it,info))
	{
		if (g->progres)
		{
			STRCopySafe(dir,g->display,MAX_MESSAGE);
			STRAppendSafe(info->name,g->display,MAX_MESSAGE);
		}
		err=IOOpenPath(info->name,OPEN_READ,dir,&in);
		if (err) { DIRClose(it); return err; }
		err=filecopy(g,in,info);
		if (err) { DIRClose(it); return err; }
	}
	DIRClose(it);
	return err;
}

Error copy(Global *g, char *filename)
{
	Channel in;
	FileInfo info;
	Error err;

	/* open the input and output files */
	catch( IOOpen(filename, OPEN_READ, &in) );
	catch( IOFileInfoGet(in,&info) );
	if (info.type==FILETYPE_DIRECTORY)
	{
		IOClose(in);
		return treecopy(g,filename,&info);
	}
	else
	{
		if (g->progres) STRCopySafe(filename,g->display,MAX_MESSAGE);
		return filecopy(g,in,&info);
	}
}

Error redraw(PWObject object)
{
	Global *g;
	PWQuery(object,PW_GLOBAL_AUXILIARY,&g);
	return PWChange(g->message,PW_INFOSTRING_TEXT,g->display,NULL);
}

Error progres(Global *g)
{
	Error err;
	PWObject win;

	catch( PWCreate(NULL,&win,PW_TYPE_OUTLINE,NULL) );
	catch( PWCreate(win,NULL,PW_TYPE_INFOSTRING,
				PW_INFOSTRING_TEXT, "now copying the file :               ",
				NULL) );
	catch( PWCreate(win,&g->message,PW_TYPE_INFOSTRING,
				PW_INFOSTRING_AUTOSIZE, FALSE,
				PW_INFOSTRING_TEXT, g->display,
				PW_GLOBAL_AUXILIARY, g,
				PW_SYSTEM_ACTION_EVENT1, redraw,
				NULL) );
	g->progres=TRUE;
	return PWActivate(win);
}
