/*
	utility program for the ProWesS loader

	COPYNAME : Copy one or more files, changing the filename in the proces.
	The file versions and update and backup dates are retained when
	copying.

	The parameters are pairs of source and destination filenames,
	and possibly some flags.

	If an error occurs while copying, then the error message is returned.
	Files which are only partly copied will then be deleted.

	possible flags :
	-move : delete all the files after copying
	-backup : only copy files if the distination file (if exists) is older than
		the source
	-progres, -progress : display progres report
	-new : only copy new files, do not overwrite existing files

	written by Joachim Van der Auwera for PROGS
	(c) 1996 by PROGS, Professional & Graphical Software
	started 24 April 1996 based on copy
	23.07.1996 sometimes ERR_MCHK is done in IOFileLoad but not in IOGetBytes
			   so use slower method if fast one fails
	02.09.1996 -new option added
	04.03.1997 -new should now function properly
*/

#include "str.h"
#include "mem.h"
#include "dir.h"
#include "io.h"
#include "job.h"
#include "ProWesS.h"

#define catch(x) if (err=(x)) return err; else

typedef struct {
	char **params;					/* parameters */
	char move;						/* move files */
	char backup;					/* backup files */
	char new;						/* only new files */
	char progres;					/* display progres of copying */
	Size bufflen;					/* length of buffer for copying */
	char *buffer;					/* base of buffer or NULL */
	Job subjob; 					/* subjob for progres report window */
	char *fromname,*toname; 		/* filenames for copying */
	PWObject mess1,mess2;			/* object with message */
	} Global;

Error filecopy(Global *g);
Error progres(Global *g);

Error init(char *param)
{
	Error err;
	char **params;
	int count=0;					/* no of files to copy */
	Global g;

	MEMClear(sizeof(Global),&g);

	/* parse the command line parameters */
	catch( STRBreak(param,&g.params) );
	for (params=g.params ; *params ; params++)
	{
		if (STRSameCI("-path",*params))
		{
			if (!params[1]) return ERR_IPAR;
			*params++=NULL;
			*params  =NULL;
		}
		else
		if (STRSameCI("-move",*params))
		{
			g.move=TRUE;
			*params=NULL;
		}
		else
		if (STRSameCI("-backup",*params))
		{
			g.backup=TRUE;
			*params=NULL;
		}
		else
		if (STRSameCI("-progres",*params) || STRSameCI("-progress",*params))
		{
			g.progres=TRUE;
			*params=NULL;
		}
		else
		if (STRSameCI("-new",*params))
		{
			g.new=TRUE;
			*params=NULL;
		}
		else
			count++;
	}
	if (g.new) g.backup=FALSE;

	/* display progress report window */
	if (g.progres)
	{
		g.progres=FALSE;
		JOBCall(progres,"copy",64,DEFAULT_STACK,&g.subjob,1,&g);
	}

	/* params should indicate first file/directory to copy */
	params=g.params;
	while (!*params) params++;

	for ( ; count>=2 ; params++, count-=2)
	{
		while (!*params) params++;
		g.fromname=*params++;
		while (!*params) params++;
		g.toname=*params;

		if (g.progres) JOBEventSend(g.subjob,EVENT_LEVEL1);

		err=filecopy(&g);
		if (err)
		{
			/* delete destination */
			IODelete(g.toname);
			return err;
		}
	}
	return ERR_OK;
}

Error filecopy(Global *g)
{
	Channel in,out;
	Error err;
	FileInfo info;
	Size done, todo;

	/* open input file and read file header */
	catch( IOOpen(g->fromname, OPEN_READ, &in) );
	catch( IOFileInfoGet(in,&info) );

	/* open the output file */
	if (g->backup)
	{
		err=IOOpen(g->toname,OPEN_OLD,&out);
		if (!err)
		{
			Time old;
			IOUpdateTimeGet(out,&old);
			if (info.UpdateTime<=old)
			{
				IOClose(out);
				IOClose(in);
				return err;
			}
			IOPositionSet(out,0);
			IOTruncate(out);
		}
		else
		{
			err=IOOpen(g->toname,OPEN_OVER,&out);
			if (err) { IOClose(in); return err; }
		}
	}
	else
	{
		err=IOOpen(g->toname,g->new?OPEN_NEW:OPEN_OVER,&out);
		if (err) { IOClose(in); return (err==ERR_FEX && g->new)?ERR_OK:err; }
	}

	/* allocate the file buffer */
	if (info.length>g->bufflen)
	{
		Size size=(info.length+0xfff)&0xfffff000;

		MEMRelease(g->buffer);
		err=MEMAllocate(size,&g->buffer);
		while (err)
		{
			size=size/2;
			err=MEMAllocate(size,&g->buffer);
		}
		g->bufflen=size;
	}

	/* read and write all the data */
	if (info.length<=g->bufflen)
	{
		err=IOFileLoad(in,info.length,g->buffer);
		if (err)
		{
			/* if medium check, try again the slow way */
			/* this might be more accurate (proven in some tests) */
			if (err==ERR_MCHK) goto slow;
			IOClose(out); IOClose(in); return err;
		}
		err=IOFileSave(out,info.length,g->buffer);
		if (err) { IOClose(out); IOClose(in); return err; }
	}
	else
	{
slow:	for (todo=info.length ; todo>0 ; todo-=done)
		{
			err=IOGetBytes(in,min(todo,g->bufflen),g->buffer,&done);
			if (err) { IOClose(out); IOClose(in); return err; }
			err=IOPutBytes(out,done,g->buffer,NULL);
			if (err) { IOClose(out); IOClose(in); return err; }
		}
	}

	/* fix the header */
	IOFileInfoSet(out,&info);

	/* copying this file was succesful, so should not be deleted */
	g->toname="";

	if (g->move)
	{
		IOClose(in);
		IODelete(g->fromname);
	}

	/* done */
	IOClose(out);
	IOClose(in);
	return err;
}

Error redraw(PWObject object)
{
	Global *g;
	PWQuery(object,PW_GLOBAL_AUXILIARY,&g);
	PWChange(g->mess1,PW_INFOSTRING_TEXT,g->fromname,NULL);
	return PWChange(g->mess2,PW_INFOSTRING_TEXT,g->toname,NULL);
}

Error progres(Global *g)
{
	Error err;
	PWObject win;

	catch( PWCreate(NULL,&win,PW_TYPE_OUTLINE,NULL) );
	catch( PWCreate(win,NULL,PW_TYPE_INFOSTRING,
				PW_INFOSTRING_TEXT, "now copying the file with new name : ",
				NULL) );
	catch( PWCreate(win,&g->mess1,PW_TYPE_INFOSTRING,
				PW_INFOSTRING_AUTOSIZE, FALSE,
				PW_INFOSTRING_TEXT, " ",
				NULL) );
	catch( PWCreate(win,&g->mess2,PW_TYPE_INFOSTRING,
				PW_INFOSTRING_AUTOSIZE, FALSE,
				PW_INFOSTRING_TEXT, " ",
				PW_GLOBAL_AUXILIARY, g,
				PW_SYSTEM_ACTION_EVENT1, redraw,
				NULL) );
	g->progres=TRUE;
	return PWActivate(win);
}
