/*
	utility program for ProWesS, the ProWesS loader

	The program has one parameter : the startup (or loader) file. Startup files
	are either called "startup" or have an extension of "_ldr".

	The search path for searching files is the directory where the startup file
	was, possibly followed by a user defined searchpath (see '+'). The
	searchpath is passed to all programs which are executed by the ProWesS
	loader, preceded by "-path".

	Startup files have the following format :

	- lines containing "PASSPATH" or "NOPASSPATH" (case dependent) are
	instructions for the ProWesS loader itself and indicate whether the current
	path should be passed in the command line or not. The default is
	"PASSPATH"

	- empty lines and lines starting with a semicolon or percentage character
	(";" or "%") are discarded as comments. Lines starting with "%%" are
	comments which are also used by maintenance software.

	- line starting with '&', followed by a program name (not separated), and
	then the parameters : execute thing or file with that name and wait for the
	result (even the program name can contain spaces when enclosed in double
	quotes).

	- line starting with '*', followed by the name of a startup file (not
	separated) : run a different startup file (similar to include, except that
	the search path is redetermined (and restored afterwards).

	- line starting with '+', followed by a searchpath will replace the
	searchpath by the directory of the startup file, followed by the searchpath
	which is given as parameter.

	- line starting with a program name (not separated), and then the
	parameters : execute thing or file with that name and wait for the result
	(even the program name can contain spaces when enclosed in double quotes).

	- It is possible to split a line over several physical lines by inserting
	a backslash just before the newline.

	- on the (first) line of a command, the leading spaces are discarded.

	written by Joachim Van der Auwera for PROGS
	(c) 1996 by PROGS, Professional & Graphical Software
	started 07 January 1996
	30.03.1996 now -path option is accepted (and used) to search file
	16.04.1996 - if loader file is an executable, then that file is executed
			   (owned by the system).
			   - if loader file has an invalid filetype, then ERR_IPAR
			   - line length no longer limited, and lines can now be split.
			   However the pathname is still limited to 120 chars.
	05.07.1996 if filename can't be loaded, try to call an executable thing
			   with that name
	14.07.1996 leading spaces are now discarded, this allows lines with only
			   spaces to be recognized as being empty
	08.08.1996 PASSPATH and NOPASSPATH added
*/

#include "sms_h"
#include "io_h"
#include "str_h"
#include "job_h"
#include "thing_h"

#define catch(x) if (err=(x)) return err; else
#define MAX_PATH_LENGTH 120+IO_MAXDIRECTORY

Error load(char *filename, char *ppath);

Error init(char *param)
{
	char *filename=NULL, *path=PATH_DATAPROG;
	char **params;
	int i; Error err;

	/* parse the command line parameters */
	catch( STRBreak(param,&params) );
	for (i=0 ; params[i] ; i++)
	{
		if (STRSameCI("-path",params[i]))
		{
			if (params[++i])
				path=params[i];
			else
				return ERR_IPAR;
		}
		else
			filename=params[i];
	}
	if (!filename) return ERR_IPAR;

	return load(filename, path);
}

Error load(char *filename, char *ppath)
{
	Error err;
	Channel file;
	char *pwroot;
	char path[MAX_PATH_LENGTH];
	char orgpath[IO_MAXDIRECTORY];
	char **params;
	char wait;
	int count;
	char *prgname;
	Size len, todo;
	FileType ft;
	char *base, *parse;
	char passpath=TRUE;

	/* open the input file if any */
	err=IOOpenPath(filename,OPEN_READ,ppath,&file);
	if (err) return THINGExec(filename,8,SYSTEM,FALSE,0,NULL,0,NULL,NULL);
	
	/* check input file, just execute if it was an executable */
	err=IOFileTypeGet(file,&ft);
	if (err) { IOClose(file); return err; }
	if (ft==FILETYPE_EXECUTABLE)
	{
		IOClose(file);
		return JOBLoad(filename,ppath,8,SYSTEM,FALSE,0,NULL,0,NULL,NULL);
	}
	if (ft!=FILETYPE_TEXT) return ERR_IPAR;

	/* build the searchpath for this file */
	err=IOName(file,NULL,orgpath);
	if (err) { IOClose(file); return err; }
	STRCopy(orgpath,path);
	STRAppend(";",orgpath); 		/* prepare for later use */

	/* allocate and load this file */
	err=IOLength(file, &len);
	if (len==0 || err) { IOClose(file); return err; }
	err=MEMAllocate(len+3, &base);
	if (err) { IOClose(file); return err; }
	err=IOFileLoad(file,len,base+2);
	IOClose(file);
	if (err) return err;

	/* process, all \n should be translated to \0, except when preceded by */
	/* a backslash. In that case the backslash has to be removed. */
	todo=len+1; parse=base+2;
	while (todo>0)
	{
		while (todo-- && *parse++!='\n');
		if (todo)
		{
			if (parse[-2]!='\\') parse[-1]='\0';
			else  parse[-2]=' ', parse[-1]=' ';
		}
	}

	/* process the lines in this file */
	todo=len; parse=base+2;
	while (todo>0)
	{
		while (*parse==' ') parse++;
		switch(*parse)
		{
			case '\0': case ';': case '%':
				/* comment line */
				break;
			case '&':
				wait=TRUE;
				parse++; todo--;
				goto execute;
			case '*':
				err=load(parse+1,path);
				if (err) { MEMRelease(base); return err; }
				break;
			case '+':
				STRCopy(orgpath,path);
				STRAppendSafe(parse+1,path,MAX_PATH_LENGTH);
				break;
			default:
				if (STRSameCD(parse,"PASSPATH"))
				{
					passpath=TRUE;
					break;
				}
				if (STRSameCD(parse,"NOPASSPATH"))
				{
					passpath=FALSE;
					break;
				}
				wait=FALSE;
execute:		parse[-2]='a'; parse[-1]=' ';
				err=STRBreak(parse-2,&params);
				if (err) { MEMRelease(base); return err; }
				if (!(prgname=params[1]))
				{
					MEMRelease(params);
					MEMRelease(base);
					return err;
				}
				params[0]="-path";
				params[1]=path;
				for (count=2 ; params[count] ; count++);
				err=THINGExec(prgname,8,wait?MYSELF:SYSTEM,wait,
								passpath?count:count-2,
								passpath?params:params+2,
								0,NULL,NULL);
				if (err)
					err=JOBLoad(prgname,path,8,wait?MYSELF:SYSTEM,wait,
								passpath?count:count-2,
								passpath?params:params+2,
								0,NULL,NULL);
				MEMRelease(params);
				if (err) { MEMRelease(base); return err; }
		}
		/* go to start of next line */
		STRLength(parse,&len);
		todo-=len+1;
		parse+=len+1;
	}
	err=ERR_OK;
	MEMRelease(base);
	return err;
}
